"""
  Reference checker for Silva.

"""

# Zope
from OFS import SimpleItem, PropertyManager
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo, Permissions
from Products.PythonScripts.Utility import allow_class
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from DateTime import DateTime
from Acquisition import aq_base, aq_inner
import zLOG

from Products.Silva import SilvaPermissions
from Products.Silva.helpers import add_and_edit
try:
    from Products.Silva.interfaces import IContent, IContainer, IPublishable
except ImportError:
    from Products.Silva.IContent import IContent
    from Products.Silva.IContainer import IContainer
    from Products.Silva.IPublishable import IPublishable


# Python
import sys
from types import StringType, UnicodeType, TupleType, ListType
from urlparse import urlparse, urlunparse
from urllib import urlopen, quote_plus
import re
from time import sleep
from thread import allocate_lock
import time


def dummy_setdefaulttimeout(timeout):
    zLOG.LOG('Silva Reference Checker',-100,
             'setting the default timeout to %s is ignored' % timeout)

def dummy_getdefaulttimeout():
    return -1.0

try:
    from socket import setdefaulttimeout, getdefaulttimeout
except ImportError:
    setdefaulttimeout = dummy_setdefaulttimeout
    getdefaulttimeout = dummy_getdefaulttimeout

uncheck_list = []
try:
    from socket import ssl
    del ssl
    # ssl support ok
except ImportError:
    uncheck_list.append('^https://.*')


# dummy check for external data sources, if nothing found
def IExtSource_isImplementedBy(object):
    return 1

try:
    from Products.SilvaExternalSources.interfaces import IExternalSource
    IExtSource_isImplementedBy = IExternalSource.isImplementedBy
except ImportError:
    pass

try:
    from Products.Silva.ISQLDataSource import ISQLDataSource
    IExtSource_isImplementedBy = ISQLDataSource.isImplementedBy
except ImportError:
    pass

# local
from ThreadPool import LoneThread, ThreadPool


#
# use a global lock for reference checking here
# this means only one external reference check may
# run at the same time. Otherwise concurrent reference
# checks on the _same_ checker caused obscure memory leaks.
# Unfortunately this global lock prevents concurrent
# checks on _different_ checkers, too.
# Doesn't use threading.Lock, as this seems to bug with 2.1.3 (?)
#
_global_lock  = allocate_lock()
thread_pool = None

class ReferenceInfo:

    # codes for the reference type

    # unknown / unclassified link
    REF_UNKNOWN = 0 
    # an internal link to something
    REF_INTERN = 1
    # an internal reference to something
    REF_MAILTO = 2
    # an external link
    REF_EXTERN = 3

    REF = {
        REF_UNKNOWN: 'Unknown',
        REF_INTERN: 'Intern',
        REF_MAILTO: 'Mailto:',
        REF_EXTERN: 'Extern'
        }


    # codes for restrictions
    RESTRICT_NONE=0
    RESTRICT_SILVA_OBJECT = 1
    RESTRICT_IMAGE = 2
    RESTRICT_FILE = 3
    RESTRICT_CONTENT = 4
    RESTRICT_FOLDER = 5
    # hmmm, more ?
    RESTRICT_CODE = 6
    RESTRICT_EXTSOURCE = 7

    RESTRICT = {
        RESTRICT_NONE: 'None',
        RESTRICT_SILVA_OBJECT: 'Silva Object',
        RESTRICT_IMAGE: 'Silva Image',
        RESTRICT_FILE: 'Silva File',
        RESTRICT_CONTENT: 'Silva Content',
        RESTRICT_FOLDER: 'Silva Folder',
        RESTRICT_CODE: 'Code Source',
        RESTRICT_EXTSOURCE: 'Silva External Source',
        }


    # information for path completion
    # (internal references only)
    
    # to be determined; checker will use default
    NO_PATH = 0
    # resolve path from zope root
    PATH_ZOPE_ROOT = 1
    # resolve path from silva root
    PATH_SILVA_ROOT = 2
    # resolve path from outmost container of containing object
    PATH_CONTAINER = 3
    # resolve path from containing object
    PATH_OBJECT = 4
    # resolve from the container _this_ object is located in
    PATH_THIS_CONTAINER = 5

    PATH = {
        NO_PATH: 'Undetermined',
        PATH_ZOPE_ROOT: 'Zope Root',
        PATH_SILVA_ROOT: 'Silva Root',
        PATH_CONTAINER: 'Current Container',
        PATH_OBJECT: 'Current Object',
        PATH_THIS_CONTAINER: 'This Container',
        }

    # status codes of the reference

    # unclassified
    STATUS_UNKNOWN = -1
    # ok
    STATUS_OK = 0
    # reference missing
    STATUS_MISSING = 1
    # reference exists but has no public version
    STATUS_UNPUBLISHED = 2
    # intern reference points to wrong type
    STATUS_WRONG_TYPE = 3
    # weird acquisition: the reference points to an object
    # in a non-straigth way
    STATUS_WEIRD_ACQUISITION = 4
    # external link has timeout
    STATUS_TIMEOUT = 6
    # external links which result in a error code
    STATUS_BROKEN = 7
    # external links which result in a non-200 code (e.g. redirects)
    STATUS_WARNING = 8

    # unchecked by config
    STATUS_UNCHECKED = 10

    STATUS = {
        STATUS_UNKNOWN: 'Unknown',
        STATUS_OK: 'OK',
        STATUS_MISSING: 'Missing',
        STATUS_WRONG_TYPE: 'Wrong type',
        STATUS_WEIRD_ACQUISITION: 'Weird',
        STATUS_TIMEOUT: 'Timeout',
        STATUS_UNCHECKED: 'Unchecked',
        STATUS_UNPUBLISHED: 'Unpublished',
        STATUS_BROKEN: 'Error', # FIXME: wrong code        
        STATUS_WARNING: 'Warning', # FIXME: better message?
        }

    # hm :(
    __allow_access_to_unprotected_subobjects__=1

    def __init__(self, obj_path, target_path, ref_type=REF_UNKNOWN, status=STATUS_UNKNOWN,
                 restriction=RESTRICT_NONE, path_info=NO_PATH):
        if type(obj_path) != type( () ):
            obj_path = obj_path.getPhysicalPath()
        self.object = obj_path
        self.target = target_path # usually a string, but sometimes a tuple of ids (_content_path of Ghosts)
        self.ref_type = ref_type
        self.restriction = restriction
        self.status = status
        self.path_info = path_info
        self.message = ''

    def get_target(self, withAnchor=1):
        """Return target as string. Remove anchor if desired."""
        target = self.target
        if type(target) == TupleType or type(target) == ListType:
            target =  '/'.join(self.target)
        if type(target) == UnicodeType:
            target = str(target)        
        if not withAnchor:
            try:
                # remove anchor from target
                res = target.split('#')
                target = res[0]
            except: pass
        return target
        

    def set_message(self, msg):
        self.message=msg

    def set_status(self, status):
        self.status=status

    def refinfo(self):
        "onestring refinfo for debugging"
        return "type=%s restrict=%s path=%s status=%s" %\
               (self.REF.get(self.ref_type),
                self.RESTRICT.get(self.restriction),
                self.PATH.get(self.path_info),
                self.STATUS.get(self.status)
                )
    
    def get_problem_description(self):
        """used by the user interface for getting the problem description"""
        return self.STATUS.get(self.status,'[undefined]')


    def __str__(self):
        return "<RefInfo: %s object=%s target=%s>" % \
               (self.refinfo(),
                self.object,
                self.get_target() or '[[empty]]')
    
    __repr__ = __str__

    # define equality for the tests
    def __eq__(self, other):
        return self.object == other.object and \
               self.get_target() == other.get_target() and \
               self.ref_type == other.ref_type and \
               self.restriction == other.restriction and \
               self.status == other.status and \
               self.path_info == other.path_info

    def __ne__(self, other):
        return not self.__eq__(other)

    # no __hash__ : should not be used as dictionary key either ...


#comparision methods

def cmp_ReferenceInfo_by_object(ref1, ref2):
    """ compare two reference infos first by their object,
    second by their status, and finally by their target.
    """
    # different object may still compare to be equal here ...
    res = cmp(ref1.object, ref2.object)
    if res != 0:
        return res
    res = cmp(ref1.status, ref2.status)
    if res != 0:
        return res
    return cmp(ref1.get_target(), ref2.get_target())

def cmp_ReferenceInfo_by_status(ref1, ref2):
    """ compare two reference infos first by their status,
    second by their object, and finally by their target
    """
    # different object may still compare to be equal here ...
    res = cmp(ref1.status, ref2.status)
    if res != 0:
        return res
    res = cmp(ref1.object, ref2.object)
    if res != 0:
        return res
    return cmp(ref1.get_target(), ref2.get_target())

allow_class(ReferenceInfo)


class ReferenceList(SimpleItem.SimpleItem):
    """put reference checking data in separate objects to avoid database conflicts
    when internal/external checkers run at the same time"""

    meta_type = 'Dummy List for Reference Checker'

    def __init__(self):
        self._list = []
        self._last_check = None

InitializeClass(ReferenceList)


class ReferenceChecker(SimpleItem.SimpleItem):

    meta_type = 'Silva Reference Checker'

    security = ClassSecurityInfo()

    manage_options = (
        ({'label':'Internal References', 'action':'manage_references'},
         {'label':'External References', 'action':'manage_linkcheck'},
         {'label':'Configuration', 'action':'manage_config'}) + \
        ()
        )


    # matching references will always receive status unchecked
    _never_check_strings = uncheck_list

    # matching references will always be considered external
    _classify_external_strings = []

    # path mismatches will not be considered weird for these objects
    # XXX: obscure; needed for accessing objects in the 'edit' and 'public' attributes
    _strange_meta_types = ['Filesystem Page Template', 'Filesystem Script (Python)']


    # used link checker.
    # You may change this value to introduce Your custom link checker
    # either by monkey patching this value, on ovberriding it
    # in a derived class of the link checker.
    #
    #  The value of this variable should be:
    #    * something callable, accepting the refinfo as one argument
    #    * when called, returning something callable which checks the url
    #
    #  The later method should by no means raise an exception, or the
    # reference checker will block.
    #
    # The most prominent candidate for this variable is a class, whose
    # instances implement a __call__ method, just in case You wonder, why
    # the requirements are that odd looking.

    # XXX "late import" to break cycle in the import statements
    import link_checker
    linkchecker_factory = link_checker.LinkCheck


    def __init__(self, id):
        self.id = id
        self.title = self.meta_type
        # all changing data in ReferenceList() so that external
        # and internal checking can run in parallel
        self._external_references = ReferenceList()
        self._internal_references = ReferenceList()
        # compiled regex, strings
        self._never_check_patterns, self._never_check_strings = \
                                    self._regex_helper(self._never_check_strings)
        self._classify_external_patterns, self._classify_external_strings = \
                                    self._regex_helper(self._classify_external_strings)
        # show links not checked by config
        self._show_unchecked = 1
        # show references with warning state
        self._show_warnings = 0
        # default timeout for external links
        self._timeout = 10.0
        # parallel threads checking external links
        self._thread_count = 10

        # default prefix for absolute-path references
        # (i.e. no protocol, but starting with a "/" ; see RFC 2396, cp 5)
        self._default_abs_root = ReferenceInfo.PATH_SILVA_ROOT
        
        # default prefix for relative-path references
        # (i.e. no protocol, not starting with a "/")
        self._default_rel_root = ReferenceInfo.PATH_CONTAINER


    #
    # helpers to look up objects
    #

    def _get_target_context(self, ref):
        # return the context to look up the target
        path_info = ref.path_info
        if path_info == ref.PATH_ZOPE_ROOT:
            context = self.getPhysicalRoot()
        elif path_info == ref.PATH_SILVA_ROOT:
            context = self.get_root()
        elif path_info == ref.PATH_CONTAINER:
            context = self._get_referring_object(ref).get_container()
        elif path_info == ref.PATH_OBJECT:
            context = self._get_referring_object(ref)
        elif path_info == ref.PATH_THIS_CONTAINER:
            context = self.aq_inner.aq_parent
        else:
            # gosh: unknown or undetermined path info ?
            self._log(100,
                     'cannot find context for type %s' % \
                     ref.PATH.get(path_info, path_info),
                     ref)
            return None
        return context
        
    def _get_referred_object(self, ref):
        """we have a reference to some zopish thingy: get it"""
        context = self._get_target_context(ref)
        #if context is None:
        #    raise TypeError, str(ref)
        obj = self._get_object_for_url(ref.get_target(), context)
        return obj

    def _get_referring_object(self, ref):
        """get the object containing the reference"""
        # FIXME: could this return None,
        # if object ceased to exist suddenly ?
        object = ref.object
        return self._get_object_for_url(object, self.getPhysicalRoot())


    def _get_object_for_url(self, path, context):
        """ return the object the path points to, or None.
        the path should be relative to the context; leading slashes are cut.
        """
        # hm, this hackery here makes it necessary
        # to add the path_info in the monkey for the
        # 'image' reference; is this really needed?
        if path and path[0]=='/':
            path = path[1:]
        
        object = context.unrestrictedTraverse(path, None)
        if object and not hasattr(aq_base(object), 'meta_type'):
            # object is not a zope object ?
            # well, it is possible that this is either:
            #  - a method
            #  - a class attribute without meta type
            #    (like the view attribute
            self._log(-100, 'no meta type for object at '+path)

            if hasattr(aq_base(object), 'index_html'):
                # hm, maybe its a view attribute ...
                object = object.index_html()
            # its something else: let's ignore it for now, even if it sucks & bugs
            object = None
        return object


    
    def _get_references_helper(self, object, references):
        """Walk the object tree and ask each object for references it uses."""
        refs = object.get_references()
        if refs:
            references.extend(refs)

        # recurse in case of containers
        if object.implements_container():
            # if there is a reference checker here: do not recurse
            other_refs = object.objectValues(self.meta_type)
            if other_refs and not self in other_refs:
                return
            
            default = object.get_default()
            if default is not None:
                self._get_references_helper(default, references)
            [ self._get_references_helper(obj, references) \
              for obj in object.get_ordered_publishables() ]
            
            [ self._get_references_helper(obj, references) \
              for obj in object.get_assets() ]



    def _never_check(self, ref):
        """return true if ref matches one of the never_check patterns"""
        target = ref.get_target()
        for p in self._never_check_patterns:
            if p.match(target):
                return 1
        return 0

    def _classify_external_by_pattern(self, target):
        """return true if ref matches one of the classify_external patterns"""
        for p in self._classify_external_patterns:
            if p.match(target):
                return 1
        return 0


    #
    # Determine the "type"
    #
    def _determine_reference_types(self, references):
        """Loop over the list of reference and determine the type for each."""
        for ref in references:
            type = ref.ref_type
            # config says it's external, so be it
            if self._classify_external_by_pattern(ref.get_target()):
                type = ref.REF_EXTERN
            if type == ReferenceInfo.REF_UNKNOWN:
                # find out whether it's internal or external if not already known
                type = self._determine_intern_or_extern_or_mailto(ref)

            # at this point ref types should be known
            assert type == ReferenceInfo.REF_INTERN or \
                   type == ReferenceInfo.REF_EXTERN or \
                   type == ReferenceInfo.REF_MAILTO

            ref.ref_type = type


    def _determine_intern_or_extern_or_mailto(self, ref):
        """ determine whether ref is intern or extern."""
        url = ref.get_target()
        url = url.strip() # XXX necessary? or does urlparse do it anyway?
        protocol, host, path, params, query, anchor = urlparse(url)

        if protocol == 'mailto':
            return ref.REF_MAILTO
        if not protocol:
            return ref.REF_INTERN
        else:
            return ref.REF_EXTERN


    #
    # determine the "status" (for internal references)
    #
    def _calculate_status_internal(self, ref, referring_object, target_object):
        """we have a reference to some zopish thingy: calculate status"""
        if self._test_wrong_type(ref, referring_object, target_object):
            ref.status = ref.STATUS_WRONG_TYPE
        elif self._test_is_not_published(ref, referring_object, target_object):
            ref.status = ref.STATUS_UNPUBLISHED
        elif self._test_weird_acquisition(ref, referring_object, target_object):
            ref.status = ref.STATUS_WEIRD_ACQUISITION
        else:
            ref.status = ref.STATUS_OK
            

    def _test_wrong_type(self, ref, referring_object, target_object):
        # Check for wrong type
        if ref.restriction == ref.RESTRICT_SILVA_OBJECT:
            if not self.service_view_registry.has_view('public',target_object.meta_type):
                return 1
        if ref.restriction == ref.RESTRICT_IMAGE:
            if target_object.meta_type != 'Silva Image':
                return 1
        if ref.restriction == ref.RESTRICT_FILE:
            if target_object.meta_type != 'Silva File':
                return 1
        if ref.restriction == ref.RESTRICT_CONTENT:
            if not IContent.isImplementedBy(target_object):
                return 1
        if ref.restriction == ref.RESTRICT_FOLDER:
            if not IContainer.isImplementedBy(target_object):
                return 1
        if ref.restriction == ref.RESTRICT_EXTSOURCE:
            if not IExtSource_isImplementedBy(target_object):
                return 1
        return 0
        

    def _test_weird_acquisition(self, ref, referring_object, target_object):
        # check for weird acquisition
        
        if target_object.meta_type in self._strange_meta_types:
            # ... but not for these oddities
            return 0
        
        target = ref.get_target(withAnchor=0)
        path = target.split('/')
        if path and path[-1]=='':
            # cut trailing slash
            path.pop()
        
        my_path = target_object.getPhysicalPath()

        if ref.path_info == ref.PATH_ZOPE_ROOT:
            # FIXME: assume path is an image ref or the like
            # this will bug, if the public root is equal to the Zope root.
            # as all links are assumed to be public visible
            pass
        else:
            # resolve ".." in the path
            # (otherwise links containig ".." will be reported as weird always)
            # FIXME: this should only happen for links and is unnecessary
            # for other references

            # to avoid getting below the public visible root
            # first make the reference absolute :
            context = self._get_target_context(ref)
            if path and path[0]=='':
                # cut leading slash
                path.pop(0)
            
            tmp_path = list(context.getPhysicalPath()) + path

            # ... then cut off the public visible root:
            # (XXX calculating the public visible root via a ref-info dummy is silly)
            dummy_ref = ReferenceInfo(('',),('',))
            dummy_ref.path_info = self._default_abs_root
            vh_root = self._get_target_context(dummy_ref)
            vh_root_len = len(vh_root.getPhysicalPath())

            del tmp_path[:vh_root_len]
            my_path = list(my_path)
            del my_path[:vh_root_len]

            # now eliminate ".." and "."
            path = []
            for x in tmp_path:
                if x == '..':
                    if path:
                        path.pop()
                    else:
                        # dived below public root
                        # XXX should we flag "weird" or something more precise?
                        return 1
                elif x == '.':
                    continue
                else:
                    path.append(x)            

        # compare path and my_path
        if tuple(path) != tuple(my_path):
            # self._log(-300,'%s is not %s' % (path, my_path))
            return 1
        return 0

    def _test_is_not_published(self, ref, referring_object, target_object):
        """check whether referenced object is published"""
        if IPublishable.isImplementedBy(target_object) and \
           not target_object.is_published():
            return 1
        return 0


    #
    # the methods that do the actual work
    # (well, or delegate the actual work ;-)
    #
    def _check_external_references(self):
        """Loop over the list of external reference and determine status for each."""

        self._external_references._list = []
        # get all references and determine their type
        references = []
        self._get_references_helper(self.aq_inner.aq_parent, references)
        self._determine_reference_types(references)

        current_timeout = getdefaulttimeout()
        setdefaulttimeout(self._timeout)
        try:
            # use a thread pool to check in parallel
            global thread_pool
            thread_pool = ThreadPool(self._thread_count)
            external_references = []
            for ref in references:
                if ref.ref_type == ref.REF_EXTERN:
                    ref.status = ref.STATUS_UNKNOWN
                    external_references.append(ref)
                    # user says never check, so be it
                    if self._never_check(ref):
                        ref.status = ref.STATUS_UNCHECKED
                    else:
                        thread_pool.queueTask(
                            self.linkchecker_factory(ref),
                            )
                        
            self._log(-300,'waiting for external references check to complete')

            thread_pool.joinAll()
        finally:
            setdefaulttimeout(current_timeout)


        # some states may have not bees set, e.g. becasue the
        # linkchek has been aborted
        # assume timeout when status has still not been set
        for ref in external_references:
            if ref.status == ref.STATUS_UNKNOWN:
                self._log(-200,'set %s to TIMEOUT' % (ref.get_target(),) )
                ref.status = ref.STATUS_TIMEOUT
                ref.message = 'Check aborted'

        thread_pool = None
        self._external_references._last_check = DateTime()
        self._external_references._list = external_references


    def _check_internal_ref(self, ref):
        # checks a single internal reference

        if ref.status != ref.STATUS_UNKNOWN:
            # already checked
            return
        
        # config says never check, so be it
        if self._never_check(ref):
            ref.status = ref.STATUS_UNCHECKED
            return

        if ref.path_info == ref.NO_PATH:
            # the reference does not know how it will be resolved
            # we use the defaults here:
            target = ref.get_target()
            if target[:1]=='/':
                ref.path_info = self._default_abs_root
            else:
                ref.path_info = self._default_rel_root
        
        
        # for internal references there has to be an object referred to ...
        target_object = self._get_referred_object(ref)
        if target_object:
            # now we have an internal object, do further classification and set status
            referring_object = self._get_referring_object(ref)
            self._calculate_status_internal(ref, referring_object, target_object)
        else:
            # ... otherwise the link is broken
            ref.status = ref.STATUS_MISSING
                
    
    def _check_internal_references(self):
        self._internal_references._list = []
        references = []
        self._get_references_helper(self.aq_inner.aq_parent, references)
        self._determine_reference_types(references)
        internal_references = []
        for ref in references:
            if ref.ref_type == ReferenceInfo.REF_INTERN:
                self._check_internal_ref(ref)
                internal_references.append(ref)
        self._internal_references._list = internal_references
        self._internal_references._last_check = DateTime()
    


    security.declareProtected(SilvaPermissions.ReadSilvaContent,
                              'check_references')
    def check_references(self, references, timeout=10.0):
        """ get a list of references passed from somewhere and checks them """
        start_time = time.time()
        self._determine_reference_types(references)
        for ref in references:
            if ref.status != ref.STATUS_UNKNOWN:
                # already checked
                continue
            if self._never_check(ref):
                ref.status = ref.STATUS_UNCHECKED
            elif ref.ref_type == ReferenceInfo.REF_INTERN:
                self._check_internal_ref(ref)
            elif ref.ref_type == ReferenceInfo.REF_MAILTO:
                 ref.status = ref.STATUS_OK # XXX could at least do syntax checks ?
            else:
                # a shortcut for external checking
                thread = LoneThread(self.linkchecker_factory(ref))
                thread.start()
                thread.join(timeout)
                del thread
                if ref.status == ref.STATUS_UNKNOWN:
                    ref.status = ref.STATUS_TIMEOUT
                    ref.message = 'test canceled after %.1f seconds' % \
                                  (timeout, )

            if time.time() - start_time > timeout:
                # oh, we are already overdue: bail out for now
                break
        
        return references
            

    #
    # ui helpers
    #
    def _regex_helper(self, in_string_patterns):
        """Compile regular expressions contained in in_string_patterns.

        Return a tuple of:
        - the list of successfully compiled patterns, and
        - the list of input strings, where patterns that failed to compile
          have an error message appended.

        If an input pattern already has an error message, the error message is removed,
        and the pattern is processed like other patterns."""
        compiled_patterns = []
        out_string_patterns = []
        for s in in_string_patterns:
            m = re.match("(.*) \[error:.*\]$", s)
            if m:
                # there was an error message but maybe the regex was corrected
                s = m.group(1)
            if not s:
                # an empty regex would match everything; guess this is not wanted
                continue
            try:
                compiled_patterns.append(re.compile(s))
                out_string_patterns.append(s)
            except Exception, msg:
                # in case of error, append error message
                out_string_patterns.append(
                    "%s [error: %s]" % (s, msg))
        return compiled_patterns, out_string_patterns


    def _sort_references(self, refs):
        # sort by status, target
        if self.REQUEST.form.get('sort_by','status') == 'object':
            refs.sort(cmp_ReferenceInfo_by_object)
        else:
            refs.sort(cmp_ReferenceInfo_by_status)
        return refs

    def _log(self, level, message):
        zLOG.LOG(self.meta_type, level, message)


    #
    # public access helpers for the UI
    #
    security.declareProtected(SilvaPermissions.ReadSilvaContent,
                              'filter_broken_references')
    def filter_broken_references(self, refs, with_unknown=None):
        ok_stat = [ ReferenceInfo.STATUS_OK ]
        if not self._show_unchecked:
            ok_stat.append( ReferenceInfo.STATUS_UNCHECKED )
        if not self._show_warnings:
            ok_stat.append( ReferenceInfo.STATUS_WARNING )
        if not with_unknown:
            ok_stat.append( ReferenceInfo.STATUS_UNKNOWN )
        return [ ref for ref in refs if
                 ref.status not in ok_stat ]
    
    security.declareProtected(Permissions.view_management_screens,
                              'get_broken_internal_references')
    def get_broken_internal_references(self):
        return self._sort_references(\
            self.filter_broken_references(self._internal_references._list,with_unknown=1))

    security.declareProtected(Permissions.view_management_screens,
                              'get_broken_external_references')
    def get_broken_external_references(self):
        return self._sort_references(\
            self.filter_broken_references(self._external_references._list,with_unknown=1))

    security.declareProtected(Permissions.view_management_screens,
                              'get_last_internal_check')
    def get_last_internal_check(self):
        dt = self._internal_references._last_check
        if dt:
            return "%02d %s %04d %02d:%02d" % (dt.day(), dt.aMonth().lower(), dt.year(), dt.hour(), dt.minute())
        else:
            return "Never checked"

    security.declareProtected(Permissions.view_management_screens,
                              'get_last_external_check')
    def get_last_external_check(self):
        dt = self._external_references._last_check
        if dt:
            return "%02d %s %04d %02d:%02d" % (dt.day(), dt.aMonth().lower(), dt.year(), dt.hour(), dt.minute())
        else:
            return "Never checked"


    security.declareProtected(SilvaPermissions.ReadSilvaContent,
                              'get_object_url')
    def get_object_url(self, ref,  human_readable=None):
        obj = self._get_referring_object(ref)
        if not human_readable:
            return obj.absolute_url()
        else:
            # display for public; allow breaks ... just hacked
            url = '/%s' % obj.absolute_url(1)
            if len(url)>60:
                url = url[:1] + url[1:].replace('/','/ ')
            return url
    
    security.declareProtected(SilvaPermissions.ReadSilvaContent,
                              'get_target_url')
    def get_target_url(self, ref, human_readable=None):
        target = ref.get_target()

        url = target
        if not human_readable:
            if ref.ref_type == ref.REF_INTERN:
                # FIXME: avoid "//" in the url
                # FIXME: what if context is None (bogus path_info) ?
                context_url = self._get_target_context(ref).absolute_url()
                if (context_url and context_url[-1]=='/'):
                    context_url = context_url[:-1]
                if (target and target[0]=='/'):
                    target = target[1:]
                url = '%s/%s' % (context_url, target)
        else:
            # display for public; allow breaks ... just hacked
            url = url.replace('?','? ', 1)
            url = url.replace('&','& ')
            if len(url)>60 and url.count('?')==0:
                url = url[:8] + url[8:].replace('/','/ ')
        return url
        
    security.declareProtected(Permissions.view_management_screens,
                              'getProperties')
    def getProperties(self):
        if setdefaulttimeout==dummy_setdefaulttimeout:
            timeout = None
        else:
            timeout = str(self._timeout)
        return {
            'show_unchecked': self._show_unchecked,
            'show_warnings': self._show_warnings,
            'never_check': self._never_check_strings,
            'classify_external': self._classify_external_strings,
            'threads': self._thread_count,
            'timeout': timeout,
            'default_abs_root' : self._default_abs_root,
            'default_rel_root': self._default_rel_root,
            }

    


    security.declareProtected(Permissions.view_management_screens,
                              'get_check_is_running')
    def get_check_is_running(self):
        if _global_lock.acquire(0):
            _global_lock.release()
            return 0
        return 1

    #
    # manipulators
    #

    security.declareProtected(Permissions.view_management_screens,
                              'setProperties')
    def setProperties(self, show_unchecked=0, show_warnings=0, never_check=(),
                      classify_external=(), timeout=None, threads=None, default_abs_root=None, default_rel_root=None, RESPONSE=None):
        """save properties"""
        messages = []
        self._show_unchecked = not not show_unchecked
        self._show_warnings = not not show_warnings
        self._never_check_strings = [str(s).strip() for s in never_check if s]
        self._classify_external_strings = [str(s).strip() for s in classify_external if s]

        if default_abs_root is not None:
            self._default_abs_root = default_abs_root
        if default_rel_root is not None:
            self._default_rel_root = default_rel_root
            
        self._never_check_patterns, self._never_check_strings = \
                                    self._regex_helper(self._never_check_strings)
        self._classify_external_patterns, self._classify_external_strings = \
                                    self._regex_helper(self._classify_external_strings)

        if threads is not None:
            try:
                threads = int(threads)
                if threads <=0:
                    raise ValueError, threads
                self._thread_count = threads
            except ValueError:
                messages.append('Number of threads should be a positive integer, not "%s".' %\
                                (threads,))

        if setdefaulttimeout != dummy_setdefaulttimeout:
            try:
                timeout = float(timeout)
                if timeout <= 0.0:
                    messages.append('Timeout should be a positive number, not "%s".' %\
                                    (timeout,))
                else:
                    self._timeout = timeout
            except ValueError:
                messages.append('Timeout is not a number: %s' % (timeout,) )


        if not messages:
            messages.append('Configuration changed.')
        if RESPONSE is not None:
            RESPONSE.redirect(
                '%s/manage_config?manage_tabs_message=%s' %
                (self.absolute_url(), quote_plus('\n'.join(messages), safe='')) )

    #
    # ZMI interface
    #

    security.declareProtected(Permissions.view_management_screens,
                              'manage_references')
    def manage_references(self):
        """ list broken internal references """
        if self.REQUEST.has_key('check'):
            self._log(-100,'start internal reference check')
            self._check_internal_references()
            self._log(-100,'end internal reference check')

        return self.manage_referencesView()

    # eeek: this can be solved less fishy
    security.declareProtected(Permissions.view_management_screens,
                              'manage_main')
    manage_main = manage_references

    security.declareProtected(Permissions.view_management_screens,
                              'manage_referencesView')
    manage_referencesView = PageTemplateFile(
        'www/viewReferences',
        globals(),  __name__='manage_referencesView')


    security.declareProtected(Permissions.view_management_screens,
                              'manage_linkcheck')
    def manage_linkcheck(self):    
        """ list broken external references """

        msg = ''

        if _global_lock.acquire(0):
            try:
                if self.REQUEST.has_key('check'):
                    self._log(-100,'start external reference check')
                    self._check_external_references()
                    self._log(-100,'end external reference check')
                    msg = 'Link check run sucessfully' 
        
            finally:
                _global_lock.release()
        else:
            if self.REQUEST.has_key('abort'):
                if thread_pool:
                    thread_pool.forceExit()
                    thread_pool.joinAll()
                    msg ="Link check has been aborted."
                    # wait a moment while the "other" thread commits the changes
                    sleep(1.0)
                    # we must redirect to ourselves,
                    # or we would get the old values here
                    self.REQUEST.RESPONSE.redirect(
                        '%s/manage_linkcheck?manage_tabs_message=%s' %
                        (self.absolute_url(), quote_plus(msg, safe='') ) )
                    return ''
            else:
                msg = 'A link check is currently running.' + \
                      '\n<!--%s\n-->' % (thread_pool and thread_pool.debug_get_state())
        
        return self.manage_linkView(message=msg)

    security.declareProtected(Permissions.view_management_screens,
                              'manage_linkView')
    manage_linkView = PageTemplateFile(
        'www/viewLinks',
        globals(),  __name__='manage_linkView')

    security.declareProtected(Permissions.view_management_screens,
                              'manage_config')
    def manage_config(self):    
        """edit configuration"""
        return self.manage_configView()

    security.declareProtected(Permissions.view_management_screens,
                              'manage_configView')
    manage_configView = PageTemplateFile(
        'www/viewConfig',
        globals(),  __name__='manage_configView')



InitializeClass(ReferenceChecker)

manage_addReferenceCheckerForm = PageTemplateFile(
    "www/addReferenceChecker", globals(),
    __name__='manage_addReferenceCheckerForm')

def manage_addReferenceChecker(self, id='service_references', REQUEST=None):
    """ add a configuration manager """
    # XXX no id check ...
    self._setObject(id, ReferenceChecker(id))
    add_and_edit(self, id, REQUEST)
    return getattr(self, id)
    
